@@ -1,7 +1,7 @@
 #! /usr/bin/perl -w
 #
-# listadmin version 2.31
-# Written 2003 - 2006 by
+# listadmin - process messages held by Mailman for approval
+# Written 2003 - 2007 by
 # Kjetil Torgrim Homme <kjetilho+listadmin@ifi.uio.no>
 #
 # Thank you, Sam Watkins and Bernie Hoeneisen, for contributions and
@@ -9,39 +9,93 @@
 #
 # Released into public domain.
 
+my $version = "2.37";
+
 use HTML::TokeParser;
 use LWP::UserAgent;
+# use LWP::Debug qw(+trace);
 use MIME::Base64;
 use MIME::QuotedPrint;
 use Data::Dumper;
 use Term::ReadLine;
-use Getopt::Std;
+use Getopt::Long;
+use Text::Reform;
+use I18N::Langinfo qw(langinfo CODESET); # appeared in Perl 5.7.2
+use Encode; # appeared in perl 5.7.1
 use strict;
 use English;
 
 my $rc = $ENV{"HOME"}."/.listadmin.ini";
 
 sub usage {
+    my ($exit_val) = @_;
     print STDERR <<_end_;
-Usage: $0 [-f CONFIGFILE] [-t MINUTES] [LISTNAME]
+Usage: $0 [-f CONFIGFILE] [-t MINUTES] [{-a|-r} FILE] [-l] [LISTNAME]
   -f CONFIGFILE    Read configuration from CONFIGFILE.
                    (default: $rc)
   -t MINUTES       Stop processing after MINUTES minutes.  Decimals are
                    allowed.
+  --mail           Turn off "nomail" flag for the specified addresses
+  --nomail         Turn on "nomail" flag for the specified addresses
+  -a FILE          Add e-mail addresses in FILE to list
+  -r FILE          Remove e-mail addresses in FILE to list
+  --add-member ADDRESS
+                   Add ADDRESS as member to list
+  --remove-member ADDRESS
+                   Remove ADDRESS from member list
+  -l               List subscribers
   LISTNAME         Only process lists with name matching LISTNAME.
+
+If options which modify members are used, LISTNAME must match exactly
+one list.
 _end_
-    exit (64);
+    exit(defined $exit_val ? $exit_val : 64);
 }
 
-my $term;
-my $ua = new LWP::UserAgent ("timeout" => 600);
+my ($opt_help, $opt_version, $opt_f, $opt_t, $opt_a, $opt_r,
+    @opt_add_member, @opt_remove_member, $opt_l);
+my $opt_mail = 1;
 
-our ($opt_f, $opt_t);
+GetOptions("help|?" => \$opt_help,
+	   "version|V" => \$opt_version,
+           "f=s" => \$opt_f,
+	   "t=i" => \$opt_t,
+	   "mail!" => \$opt_mail,
+	   "a=s" => \$opt_a,
+	   "r=s" => \$opt_r,
+	   "add-member=s" => \@opt_add_member,
+	   "remove-member=s" => \@opt_remove_member,
+	   "l" => \$opt_l)
+	or usage();
+
+usage(0) if $opt_help;
+if ($opt_version) {
+    print "listadmin version $version\n";
+    exit(0);
+}
 
-usage() unless getopts('f:t:');
 $rc = $opt_f if $opt_f;
 usage() if defined $opt_t && $opt_t !~ /\d/ && $opt_t !~ /^\d*(\.\d*)?$/;
+
+push(@opt_add_member, read_address_file($opt_a, 1)) if defined $opt_a;
+push(@opt_remove_member, read_address_file($opt_r, 1)) if defined $opt_r;
+
+my $will_modify_membership = 0;
+++$will_modify_membership if @opt_add_member;
+++$will_modify_membership if @opt_remove_member;
+
+usage() if $will_modify_membership > 1;
+usage() if defined $opt_l && $will_modify_membership;
+
+my $ua = new LWP::UserAgent("timeout" => 900, "env_proxy" => 1);
 my $time_limit = time + 60 * ($opt_t || 24*60);
+my $term;
+my $term_encoding = langinfo(CODESET());
+
+# the C and POSIX locale in Solaris uses the charset "646", but Perl
+# doesn't support it.
+$term_encoding = "ascii" if $term_encoding eq "646";
+binmode STDOUT, ":encoding($term_encoding)";
 # Turn on autoflush on STDOUT
 $| = 1;
 
@@ -66,21 +120,41 @@ if (@ARGV) {
     @lists = sort config_order keys %{$config}
 }
 
-my ($num, $count, $list, $from, $subject, $reason, $spamscore);
+if (@lists > 1 && ($will_modify_membership || defined $opt_l)) {
+    print STDERR "Too many matching lists\n";
+    print Dumper(\@lists);
+    usage();
+}
 
-format STDOUT =
+my $list = $lists[0];
 
-@<<<<<<< ========== @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
-sprintf("[%d/%d]", $num, $count), $list." "."=" x (51 - length($list))
-From:    @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
-         $from
-Subject: ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
-         $subject
-~~       ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
-         $subject
-Reason:  @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<  Spam? @<<
-         $reason,                                             $spamscore
-.
+my $subscribe_result;
+if (@opt_add_member) {
+    $subscribe_result = add_subscribers($list, $config->{$list}, $opt_mail,
+					@opt_add_member);
+}
+if (@opt_remove_member) {
+    $subscribe_result = remove_subscribers($list, $config->{$list},
+					   @opt_remove_member);
+}
+if (defined $subscribe_result) {
+    for my $addr (keys %{$subscribe_result}) {
+	print STDERR "$addr: $subscribe_result->{$addr}\n";
+    }
+    if (%{$subscribe_result}) {
+	exit(1);
+    } else {
+	print "Ok\n";
+	exit(0);
+    }
+}
+if (defined $opt_l) {
+    my @subscribers = list_subscribers($list, $config->{$list});
+    print join("\n", @subscribers, "");
+    exit(@subscribers == 0);
+}
+
+my ($num, $count, $from, $subject, $reason, $spamscore);
 
 
 for (@lists) {
@@ -125,6 +199,7 @@ for (@lists) {
     } else {
 	print "\n";
     }
+    $config->{$list}{"password"} = $pw;
 
     my %change = ();
 
@@ -140,16 +215,21 @@ for (@lists) {
 	    if ($c =~ /^\s*(\?+|h|hj?elp)\s*$/i) {
 		print <<_END_;
 Nothing will be done to the messages in the administrative queue
-unless you answer this question affirmatively.
+unless you answer this question affirmatively.  If you answer "no",
+your changes will be discarded and listadmin will proceed to the
+next mailing list.  Type "undo" to go back to the current list.
 _END_
 		goto redo_confirm;
 	    }
 	    if ($c =~ /^\s*(no?|nei|skip)\s*$/i) {
 		print "skipping ...\n";
 		next;
-	    } elsif ($c =~ /^\d$/) {
+	    } elsif ($c =~ /^\d+$/) {
 		$num = $c - 1;
 		goto restart_approval;
+	    } elsif ($c =~ /^u(ndo)?/) {
+		--$num;
+		goto restart_approval;
 	    } elsif ($c !~ /^\s*(|ja?|y|yes)\s*$/i) {
 		goto redo_confirm;
 	    }
@@ -208,7 +288,7 @@ sub process_subscriptions {
 		$change->{$id} = [ "sr", $r ];
 		last;
 	    } else {
-		print STDERR <<end;
+		print STDERR <<"_end_";
 Choose one of the following actions by typing the corresponding letter
 and pressing Return.
 
@@ -217,7 +297,7 @@ and pressing Return.
   s  Skip      -- do not decide now, leave it for later
   q  Quit      -- go on to approving messages
 
-end
+_end_
             }
 	}
     }
@@ -242,8 +322,19 @@ sub approve_messages {
     } else {
 	$dont_skip_forward = 1;
     }
-    my $prompt = 'Approve/Reject/Discard/Skip/view Body/view Full/jump #/Help/Quit';
+    my $tmpl_header = << '_end_';
+
+<<<<<<<<<<<<<<<<<<<< <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
+_end_
+    my $tmpl_message = << '_end_';
+From:     <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
+<<<<<<<<  [[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[[
+Reason:   <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<  Spam? <<<
+_end_
+
+    my $prompt = 'Approve/Reject/Discard/Skip/view Body/Full/jump #/Undo/Help/Quit';
     my @num_to_id = grep { ! /^global$/ } sort keys %{$info};
+    my @undo_list = ();
  msgloop:
     while ($num < $count) {
 	last if time > $time_limit;
@@ -252,8 +343,20 @@ sub approve_messages {
 	$subject = $info->{$id}{"subject"} || "";
 	$reason = $info->{$id}{"reason"};
 	$spamscore = $info->{$id}{"spamscore"};
-	write;
-
+	{
+	    # Get rid of warning from Encode:
+	    # "\x{516b}" does not map to iso-8859-1 at listadmin.pl line 261.
+	    # when run in non UTF-8 environment.
+	redraw:
+	    local $SIG{__WARN__} = sub {};
+	    print form({filler => {left => "=", right => "="}},
+		       $tmpl_header,
+		       "[$num/$count] =", "$list =");
+	    print form({interleave => 1},
+		       $tmpl_message,
+		       $from,
+		       "Subject:", $subject, $reason, $spamscore);
+	}
 	while (1) {
 	    my $ans;
 	    my $match = "";
@@ -287,12 +390,15 @@ sub approve_messages {
 	    $pr .= " ? ";
 	    $ans ||= prompt ($pr);
 	    $ans = "q" unless defined $ans;
-	    $ans =~ s/\s+//g;
+	    $ans =~ s/^\s+//;
+	    $ans =~ s/\s+$//;
 	    $ans = $def if $ans eq "" && defined $def;
 	    $ans = lc $ans;
 	    if ($ans eq "q") {
 		last msgloop;
 	    } elsif ($ans eq "s") {
+		# Undo will be a no-op, except it will go back to this message.
+		push(@undo_list, [$num]);
 		delete $change->{$id};
 		$dont_skip_forward = 0;
 		next msgloop;
@@ -301,9 +407,57 @@ sub approve_messages {
 		$dont_skip_forward = 1;
 		next msgloop;
 	    } elsif ($ans eq "a" || $ans eq "d") {
+		# If it is automatically discarded, add it to existing list
+		push(@undo_list, []) unless $match && @undo_list;
+		push(@{$undo_list[$#undo_list]}, $num);
 		$change->{$id} = [ $ans ];
 		$dont_skip_forward = 0;
 		last;
+	    } elsif ($ans eq "u") {
+		unless (@undo_list) {
+		    print "Nothing to undo.\n";
+		    next;
+		}
+		my @trans_list = @{pop(@undo_list)};
+		for my $m (@trans_list) {
+		    delete $change->{$num_to_id[$m - 1]};
+		}
+		$num = $trans_list[0] - 1;
+		$dont_skip_forward = 1;
+		next msgloop;
+	    } elsif ($ans =~ /^list(\s+|$)/) {
+		my @list = list_subscribers($list, $config);
+		my $member_count = scalar @list;
+		if ($POSTMATCH ne "") {
+		    @list = grep { /$POSTMATCH/ } @list;
+		    printf("Found %d matching addresses:\n  ", scalar @list);
+		} else {
+		    print "Mailing list members:\n  ";
+		}
+		print join("\n  ", @list);
+		print "\n$member_count members in total\n";
+	    } elsif ($ans =~ /^(add|nomail)(\s+|$)/) {
+		my $mail = $1 eq "add";
+		my $addr = $POSTMATCH || $from;
+		my $res = add_subscribers($list, $config, $mail, $addr);
+		for my $addr (keys %{$res}) {
+		    print "$addr: $res->{$addr}\n";
+		}
+		print "done\n";
+	    } elsif ($ans =~ /^rem(\s+|$)/) {
+		my $address = $POSTMATCH;
+		my $c = prompt ("Remove subscriber? (there is no undo!) [no] ");
+		if ($c =~ /^\s*(ja?|y|yes)\s*$/i) {
+		    print "removing...\n";
+		    my $res = remove_subscribers($list, $config, $address);
+		    for my $addr (keys %{$res}) {
+			print "$addr: $res->{$addr}\n";
+		    }
+		    print "done\n";
+		} else {
+		    print "aborted\n";
+		    next;
+		}
 	    } elsif ($ans =~ m,([/?])(.*),) {
 		my $i = $num - 1;
 		my $direction = 1;
@@ -346,37 +500,56 @@ sub approve_messages {
 		    goto redo_reject;
 		}
 
+		push(@undo_list, [ $num ]);
 		$change->{$id} = [ "r", $r ];
 		$dont_skip_forward = 0;
 		last;
 	    } elsif ($ans eq "f") {
-		print $info->{$id}{"headers"}, "\n\n", $info->{$id}{"body"};
+		# Since the raw bytes aren't really Unicode, we set
+		# the replacement sequence to be "<?>" unconditionally.
+		print degrade_charset($info->{$id}{"headers"} . "\n\n" .
+				      $info->{$id}{"body"}, "questionmark");
 	    } elsif ($ans eq "b") {
-		my $head = lc $info->{$id}{"headers"};
+		my $head = $info->{$id}{"headers"};
 		my $text = $info->{$id}{"body"};
-		if ($head =~ m,content-type:\s+text/,) {
-		    my $charset = "UNKNOWN";
-		    if ($head =~ /charset="?(iso-8859-15?|us-ascii|utf-8)"?/) {
-			$charset = $1;
-		    }
-		    if ($head =~ /content-transfer-encoding:\s+quoted-print/) {
-			$text = MIME::QuotedPrint::decode($text);
-		    } elsif ($head =~ /content-transfer-encoding:\s+base64/) {
-			$text = MIME::Base64::decode_base64($text);
-		    }
-		    $text = utf8_to_latin1 ($text) if $charset eq "utf-8";
+		my $mime_headers = "";
+		if ($head =~ m,content-type:\s*text/,i) {
+		    $mime_headers = $head;
+		} elsif ($head =~ m,content-type:\s*multipart/,i) {
+		    # This is quick and dirty, we look at the first
+		    # MIME headers in the body instead.  We can't do
+		    # proper MIME parsing since the message is
+		    # truncated by Mailman.
+		    $mime_headers = $text;
 		}
+		if ($mime_headers =~ /content-transfer-encoding:\s+(\S+)/i) {
+		    my $cte = $1;
+		    if ($cte =~ /quoted-printable/i) {
+			$text = MIME::QuotedPrint::decode($text);
+		    } elsif ($cte =~ /base64/i) {
+			# Don't bother with truncated lines.
+			$text =~ s!([A-Za-z0-9/+=]{72,76})!MIME::Base64::decode_base64($1)!ge;
+		    }
+		}
+		if ($mime_headers =~ /charset=(\S+)/i) {
+		    my $charset = $1;
+		    $charset =~ s/;$//;
+		    $charset =~ s/^"(.*)"$/$1/;
+		    $charset = guess_charset($charset, $text);
+		    eval { $text = Encode::decode($charset, $text) };
+		}
+
+		$text = degrade_charset($text, $config->{unprintable});
 		my @lines = split (/\n/, $text, 21);
 		pop @lines;
+		# local $SIG{__WARN__} = sub {}; # see comment elsewhere
 		print join ("\n", @lines), "\n";
 	    } elsif ($ans eq "t") {
 		print $info->{$id}{"date"}, "\n";
 	    } elsif ($ans eq "url") {
 		print mailman_url($list, $config->{adminurl}), "\n";
 	    } elsif ($ans eq ".") {
-		# write modifies $subject, so reinitialise it
-		$subject = $info->{$id}{"subject"} || "";
-		write;
+		goto redraw;
 	    } elsif ($ans eq "") {
 		# nothing.
 	    } else {
@@ -384,18 +557,23 @@ sub approve_messages {
 Choose one of the following actions by typing the corresponding letter
 and pressing Return.
 
-  a  Approve   -- the message will be sent to all members of the list
-  r  Reject    -- notify sender that the message was rejected
-  d  Discard   -- throw message away, don't notify sender
-  s  Skip      -- don't decide now, leave it for later
-  b  view Body -- display the first 20 lines of the message
-  f  view Full -- display the complete message, including headers
-  t  view Time -- display the date the message was sent
-  #  jump      -- jump backward or forward to message number #
-  /pattern     -- search for next message with matching From or Subject
-  ?pattern     -- search for previous message with matching From or Subject
-  .            -- redisplay entry
-  q  Quit      -- go on to the next list
+  a  Approve     -- the message will be sent to all members of the list
+  r  Reject      -- notify sender that the message was rejected
+  d  Discard     -- throw message away, don't notify sender
+  s  Skip        -- don't decide now, leave it for later
+  b  view Body   -- display the first 20 lines of the message
+  f  view Full   -- display the complete message, including headers
+  t  view Time   -- display the date the message was sent
+  #  jump        -- jump backward or forward to message number #
+  u  Undo        -- undo last approve or discard
+  /pattern       -- search for next message with matching From or Subject
+  ?pattern       -- search for previous message with matching From or Subject
+  .              -- redisplay entry
+  add [address]  -- add subscription for address (defaults to From)
+  nomail [address] -- add nomail subscription for address (defaults to From)
+  list [pattern] -- list mailing list members matching optional pattern
+  rem address    -- remove list member
+  q  Quit        -- go on to the next list
 
 end
 		print <<"end" if $listdef;
@@ -435,7 +613,7 @@ sub uio_adminurl {
 }
 
 sub mailman_url {
-    my ($list, $pattern, $params) = @_;
+    my ($list, $pattern, $params, $action) = @_;
 
     my ($lp, $domain) = split ('@', $list);
 
@@ -448,6 +626,10 @@ sub mailman_url {
     $url =~ s/\{list\}/$lp/g;
     $url =~ s/\{domain\}/$domain/g;
     $url =~ s/\{subdomain\}/$subdom/g;
+    if ($action) {
+	$url =~ s,/admindb/,/admin/,;
+	$url .= "/$action";
+    }
     $url .= "?$params" if $params;
     return $url;
 }
@@ -465,6 +647,14 @@ sub get_list {
 	return {'servererror' => $resp->status_line, 'url' => $url};
     }
     $page = $resp->content;
+    if ($page eq "") {
+	if (time - $starttime > 60) {
+	    print "Mailman server timed out?\n";
+	} else {
+	    print "empty page\n";
+	}
+	return {};
+    }
     my $dumpdir = $config->{"dumpdir"};
     if (defined $dumpdir) {
 	if (open (DUMP, ">$dumpdir/dump-$list.html")) {
@@ -611,27 +801,101 @@ sub parse_approvals {
     } until ($token->[0] eq "S" && lc ($token->[1]) eq "input");
 }
 
-# NB! lossy!
-sub utf8_to_latin1 {
-    my ($s) = @_;
-    $s =~ s/([\x80-\xff][\x80-\xbf]*)/&utf8_to_latin1_char($1)/ge;
-    return $s;
-}
+sub guess_charset {
+    my ($charset, $text) = @_;
 
-sub utf8_to_latin1_char {
-    my($first, @rest) = unpack('C*', $_[0]);
-    $first ^= 0xC2;
-    return chr($first * 0x40 + $rest[0]) if $first < 2 && @rest == 1;
-    # We simply remove the other codes, they obviously won't fit in Latin1.
-    return "";
+    # Mislabeling Shift JIS as ISO 2022 is a very common mistake.
+    if ($charset =~ /^iso-2022-jp/i && $text =~ /[\x80-\x9f]/) {
+	return "Shift_JIS";
+    }
+    return $charset;
 }
 
 sub decode_rfc2047_qp {
-    my $text = shift;
+    my ($charset, $encoded_word) = @_;
+    my $text = $encoded_word;
     $text =~ s/_/ /g;
-    return MIME::QuotedPrint::decode ($text);
+    $text = MIME::QuotedPrint::decode($text);
+    $charset = guess_charset($charset, $text);
+    eval { $text = Encode::decode($charset, $text) };
+    return defined $text ? $text : $encoded_word;
 }
 
+sub decode_rfc2047_base64 {
+    my ($charset, $encoded_word) = @_;
+    my $text = MIME::QuotedPrint::decode_base64($encoded_word);
+    $charset = guess_charset($charset, $text);
+    eval { $text = Encode::decode($charset, $text) };
+    return defined $text ? $text : $encoded_word;
+}
+
+sub decode_rfc2047 {
+    my ($hdr, $config) = @_;
+
+    # Bugs: Decodes invalid tokens, where the encoded word is
+    # concatenated with other letters, e.g.  foo=?utf-8?q?=A0=F8?=
+    # Also decodes base64 encoded words which are doubly encoded with
+    # quoted-printable.
+
+    $hdr =~ s/=\?([^? ]+)\?q\?([^? ]*)\?=/
+	    decode_rfc2047_qp($1, $2)/ieg;
+    $hdr =~ s/=\?([^? ]+)\?b\?([^? ]*)\?=/
+	    decode_rfc2047_base64($1, $2)/ieg;
+
+    return degrade_charset($hdr, $config->{unprintable});
+}
+
+sub degrade_charset {
+    my ($text, $unprintable) = @_;
+
+    # Handle unencoded Shift JIS (Japanese) text.  The input text is
+    # either raw data from the message, or Unicode, in which case it
+    # will not contain these code points.  This discrimates slightly
+    # against users of Windows-1252, which has curved quotes at 0x82
+    # (0x81 is unassigned).
+
+    if ($text =~ /[\x81\x82]/) {
+	eval { $text = Encode::decode("Shift_JIS", $text) };
+    }
+
+    # This may look a bit silly.  We first encode to the character set
+    # of our terminal.  If it is a limited character set such as
+    # Latin1, Chinese glyphs are converted into e.g. "&#x41a;", while
+    # "n with tilde" will be a single glyph.  We then convert this
+    # back to a Unicode string so that the length is right (number of
+    # glyphs, not octets) for Text::Reform.  Finally, when the Unicode
+    # string is printed to the screen, the binmode directive for
+    # STDOUT tells Perl to once more translate it into the terminal's
+    # character set.
+
+    eval {
+	$text = Encode::decode($term_encoding,
+			       Encode::encode($term_encoding, $text,
+					      Encode::FB_HTMLCREF))
+	    };
+
+    # The built-in formats for unprintable glyphs are ugly, and to be
+    # allowed to specify a code ref which returns our preferred format
+    # directly, we need to require Encode version 2.10, which feels a
+    # bit unnecessary.
+
+    if (defined $config && $unprintable eq "unicode") {
+	$text =~ s/&\#(\d+);/sprintf("<U+%04x>", $1)/ge;
+    } else {
+	$text =~ s/&\#\d+;/<?>/g;
+    }
+
+    # Get rid of ESC sequences which may cause havoc with the
+    # terminal, we only keep TAB and LF.  Also removes control
+    # characters with high bit set, 127-159, which are unallocated in
+    # Unicode.
+
+    $text =~ s/([\x00-\x08\x0b-\x1f\x7f-\x9f])/sprintf("<%02x>", ord($1))/eg;
+
+    return $text;
+}
+
+
 sub parse_approval {
     my ($mmver, $config, $parse, $data) = @_;
     my ($from, $reason, $subject, $id, $body, $headers);
@@ -660,22 +924,11 @@ sub parse_approval {
 	$parse->get_tag ("td") || die;
 	$reason = $parse->get_trimmed_text("/td");
     }
-    my $utf8 = 0;
-    # this will also decode invalid tokens, where the encoded word is
-    # concatenated with other letters, e.g.  foo=?utf-8?q?=A0=F8?=
-    $subject =~ s/=\?(us-ascii|utf-8|iso-8859-15?)\?q\?(.*?)\?=/
-	    decode_rfc2047_qp($2)/ieg;
-    $utf8 ||= 1 if defined $1 && $1 =~ /utf-8/i;
-    $subject =~ s/=\?(us-ascii|utf-8|iso-8859-15?)\?b\?(.*?)\?=/
-	MIME::Base64::decode_base64($2)/ieg;
-    $utf8 ||= 1 if defined $1 && $1 =~ /utf-8/i;
-    $subject = utf8_to_latin1 ($subject) if $utf8;
-
     $parse->get_tag ("tr") || die;	# Action:
     my $tag = $parse->get_tag ("input") || die;
     $id = $tag->[1]{"name"};
 
-    $data->{$id} = { "from" => $from,
+    $data->{$id} = { "from" => decode_rfc2047($from, $config),
 		     "subject" => $subject,
 		     "reason" => $reason };
 
@@ -725,9 +978,23 @@ sub parse_approval {
 	$body = $POSTMATCH;
 	$headers = $PREMATCH;
     }
+    $headers =~ s/\n(\s)/$1/g; # Header folding
     $headers =~ s/^\s+//;
-    $body .= "\n" unless $body =~ /\n$/;
     $data->{$id}->{"headers"} = $headers;
+
+    # Mailman decodes Subject itself, but at least version 2.0 and 2.1
+    # screw up non-ASCII characters, so we get the raw value from the
+    # headers instead.
+    if ($headers =~ /^Subject:\s*(.*)\s*$/mi) {
+	$subject = $1;
+    }
+    if ($subject =~ /[\x80-\xff]/) {
+	$subject .= " [unencoded]";
+    }
+
+    $data->{$id}->{"subject"} = decode_rfc2047($subject, $config);
+
+    $body .= "\n" unless $body =~ /\n$/;
     $data->{$id}->{"body"} = $body;
 
     return ($mmver);
@@ -770,6 +1037,7 @@ sub read_config {
     my $confirm = 1;
     my $url;
     my $spamheader;
+    my $unprintable = "questionmark";
     my %patterns = map { $_ => undef; }
                        qw (not_spam_if_from
 			   not_spam_if_subject
@@ -901,9 +1169,17 @@ sub read_config {
 			       "logfile" => $logfile,
 			       "dumpdir" => $dumpdir,
 			       "spamheader" => $spamheader,
+			       "unprintable" => $unprintable,
 			       %patterns,
 			       "order" => ++$count,
 			   };
+	} elsif ($line =~ /^unprintable\s+/) {
+	    $unprintable = unquote($POSTMATCH);
+	    unless ($unprintable =~ /^(questionmark|unicode)$/) {
+		print STDERR "$file:$lineno: Illegal format for ".
+			"unprintable characters: '$unprintable'\n";
+		exit 1;
+	    }
 	} else {
 	    print STDERR "$file:$lineno: Syntax error: '$line'\n";
 	    exit 1;
@@ -1042,18 +1318,13 @@ sub commit_changes {
 	}
 	++$changes;
 
-	# HTTP does not specify a maximum length for the URI in a GET
-	# request, but it recommends that a server does not rely on
-	# clients being able to send URIs larger than 255 octets.  the
-	# reject reason can be very long, so theoretically, we can
-	# overshoot that limit even if we change the 500 below into
-	# 250.  Mailman has been observed to reject URI's ~3400 octets
-	# long, but accept 8021.  the limit is probably based on the
-	# time taken to process, rather than the length of the URI.
-	# in times with high load on the Mailman server, it's best to
-	# keep the amount of work per request down.
+	# HTTP does not specify a maximum size for a POST request, so
+	# we could do this as one request.  However, Apache is usually
+	# set up to close the connection after the CGI script has run
+	# for 5 minutes, so we reduce the size of each request to be
+	# nice to slow servers.
 
-	if ($changes > 50) {
+	if ($changes >= 100) {
 	    $update_count += $changes;
 	    printf("sending %d updates to server, %d left    \r",
 		   $changes, $update_total - $update_count);
@@ -1083,6 +1354,234 @@ sub log_timestamp {
 		     $list, $year+1900, $mon+1, $mday, $hour, $min, $sec));
 }
 
+sub add_subscribers {
+    my ($list, $config, $mail, @addresses) = @_;
+
+    die unless @addresses;
+
+    my %params = (username => $config->{user},
+		  adminpw => $config->{password},
+		  subscribe_or_invite => 0,                # Mailman 2.x
+		  send_notifications_to_list_owner => 0,   # Mailman 2.x
+		  send_welcome_message_to_this_batch => 0, # Mailman 2.x
+		  send_welcome_msg_to_this_batch => 0,     # Mailman 1.2
+		  subscribees => join("\n", @addresses));
+    my $url = mailman_url($list, $config->{adminurl}, "", "members");
+    my $resp = $ua->post($url, \%params);
+    return $resp->status_line unless $resp->is_success;
+
+    my $result = parse_subscribe_response($resp->content);
+
+    if (!$mail) {
+	my %left = map { $_ => 1 } @addresses;
+	for my $failed (keys %{$result}) {
+	    unless ($result->{$failed} =~ /Already a member/) {
+		delete $left{$failed};
+	    }
+	}
+	@addresses = keys %left;
+    } else {
+	# We only need to reset "nomail" on the users who already were
+	# members.
+	@addresses = ();
+	for my $failed (keys %{$result}) {
+	    if ($result->{$failed} =~ /Already a member/) {
+		push(@addresses, $failed);
+	    }
+	}
+    }
+    if (@addresses) {
+	%params = (username => $config->{user},
+		   adminpw => $config->{password},
+		   user => \@addresses,
+		   setmemberopts_btn => "submit");	# Mailman 2.x
+	for my $a (@addresses) {
+	    $params{$a . "_nomail"} = "on" unless $mail;
+	    $params{$a . "_subscribed"} = "on";		# Mailman 1.2
+	}
+	$resp = $ua->post($url, \%params);
+	return $resp->status_line unless $resp->is_success;
+    }
+
+    return $result;
+}
+
+sub remove_subscribers {
+    my ($list, $config, @addresses) = @_;
+
+    my $url = mailman_url($list, $config->{adminurl}, "", "members");
+
+    # In Mailman 1.2, unsubscription happens when an address is
+    # mentioned in "user" without a corresponding
+    # "$address_subscribed" parameter
+    my %params = (username => $config->{user},
+		  adminpw => $config->{password},
+		  setmemberopts_btn => "submit",	# Mailman 2.x
+		  user => \@addresses);
+    for my $a (@addresses) {
+	$params{$a . "_unsub"} = "on";			# Mailman 2.x
+    }
+    my $resp = $ua->post($url, \%params);
+    return $resp->status_line unless $resp->is_success;
+
+    return parse_subscribe_response($resp->content);
+}
+
+
+sub parse_subscribe_response {
+    my ($page) = @_;
+
+    # Normalise, to make parsing easier (Hack!)
+    $page =~ s/<h3\>/\<h5\>/ ;
+    $page =~ s/<\/h3\>/\<\/h5\>/;
+
+    # In Mailman 1.2 and 2.0, you will not get an explicit success
+    # report when removing subscribers, so we only return the
+    # failures since the successes can be inferred anyway.
+
+    my %failure = ();
+
+    my $parse = HTML::TokeParser->new(\$page) || die;
+
+    while ($parse->get_tag ("h5")) {
+	my $h5 = $parse->get_text ("/h5");
+
+	$parse->get_tag ("ul") || die;
+	my $ul = $parse->get_text ("/ul") || die;
+
+	if ($h5 =~ /Successfully ((un)?subscribed|Removed)/i) {
+	    # hooray!
+	} elsif ($h5 =~ /Error (un)?subscribing/i) {
+	    for (split(/\n/, $ul)) {
+		chomp;
+		if (/^\s*(.*?)\s*--\s*(.*)/) {
+		    $failure{$1} = $2;
+		}
+	    }
+	} else {
+	    $ul =~ s/\n/\n\t/g;
+	    print STDERR "You have an unusual Mailman output.  Please mail ".
+		    "this message to\nkjetilho+listadmin\@ifi.uio.no:\n".
+		    "\t[$h5]\n\t[$ul]\nThanks!\n";
+	}
+	$parse->get_tag ("p") || die;
+    }
+
+    return \%failure;
+}
+
+sub list_subscribers {
+    my ($list, $config) = @_;
+
+    my $url = mailman_url($list, $config->{adminurl}, "", "members");
+    my %params = (username => $config->{user},
+		  adminpw => $config->{password},
+		  chunk => 0);
+    my $resp = $ua->post($url, \%params);
+    unless ($resp->is_success) {
+	print "$url: ", $resp->status_line, "\n";
+	return ();
+    }
+
+    my @addresses = ();
+    my ($parse, $page, $tag);
+
+ member_letter:
+    for my $letter ("a" .. "z") {
+	my $chunk = 0;
+
+	$params{chunk} = $chunk;
+
+	# Mailman 2.x specifically looks at QUERY_STRING, so chunk and
+	# letter can't be parameters to POST.  However, Mailman 1.x
+	# only looks at chunk in the POST parameters.
+	$resp = $ua->post("$url?letter=$letter&chunk=$chunk", \%params)
+		unless $letter eq "a";
+
+	while ($resp->is_success) {
+	    $page = $resp->content;
+	    $parse = HTML::TokeParser->new(\$page);
+	    my $count = 0;
+	    my $repeated = 0;
+	    my $later_letter = 0;
+	    while ($tag = $parse->get_tag("input")) {
+		my $attr = $tag->[1];
+		if ($attr->{type} =~ /^hidden$/i &&
+		    $attr->{name} =~ /^user$/i) {
+		    ++$count;
+		    my $address = $attr->{value};
+		    unless ($address =~ /\@/) {
+			# Mailman 2.x adds URL-encoding
+			$address =~ s/%([0-9a-fA-F]{2})/sprintf("%c", hex($1))/ge;
+		    }
+		    ++$later_letter if lc(substr($address, 0, 1)) gt $letter;
+		    if (grep { $_ eq $address } @addresses) {
+			++$repeated;
+		    } else {
+			push(@addresses, $address);
+		    }
+		}
+	    }
+	    last if $count == 0;
+
+	    # In Mailman 1.x, "letter" is a no-op, so $later_letter
+	    # will ~always be true and should be ignored.  Increase
+	    # chunk until we see repeats.
+
+	    # In Mailman 2.x, we need to iterate through both letter
+	    # and chunk, but if the list has few members, they will
+	    # all be listed and letter and chunk are ignored.  Also,
+	    # if there are no members for a given letter, the whole
+	    # list will be returned.
+
+	    if ($repeated) {
+		last member_letter if $later_letter;
+		next member_letter;
+	    }
+
+	    # The maximum number of addresses on each page can be
+	    # configured, by default it is set to 30, but it could in
+	    # theory be less.  To save time, we assume that we have
+	    # all the members if we got less than 20 addresses.
+	    next member_letter if $count < 20;
+
+	    ++$chunk; $params{chunk} = $chunk;
+	    $resp = $ua->post("$url?letter=$letter&chunk=$chunk", \%params);
+	}
+    }
+    return @addresses;
+}
+
+sub remove_matching_subscribers {
+    my ($list, $config, $pattern) = @_;
+    my @addresses = list_subscribers($list, $config);
+    if (defined($pattern) and $pattern ne "") {
+	@addresses = grep { /$pattern/ } @addresses;
+    }
+    my $msg = remove_subscribers($list, $config, @addresses);
+    if ($msg eq "OK") {
+	print "Removed:\n  ", join("\n  ", @addresses), "\n";
+    } else {
+	print $msg, "\n";
+    }
+}
+
+sub read_address_file {
+    my ($file, $assert_nonempty) = @_;
+    my @list = ();
+    open(F, $file) || die "$file: $!\n";
+    while (<F>) {
+	s/(^|\s)\#.*//;
+	s/^\s+//;
+	s/\s+$//;
+	next if /^$/;
+	push(@list, $_);
+    }
+
+    die "$file: no lines, aborting\n" if $assert_nonempty && @list == 0;
+    return @list;
+}
+
 sub submit_http {
     my ($url, $params, $log, $logfile) = @_;
 
@@ -1090,7 +1589,9 @@ sub submit_http {
     if ($logfile) {
 	if (open (LOG, ">>$logfile")) {
 	    LOG->autoflush(1);
+	    binmode LOG, ":encoding(" . langinfo(CODESET()) . ")";
 	    $opened = 1;
+	    local $SIG{__WARN__} = sub {}; # see comment elsewhere
 	    print LOG $log;
 	} else {
 	    print STDERR "WARNING: Failed to append to $logfile: $!\n";
